import time
from typing import Dict
from .....com.orbitgis.tool.message.Message import Message
from .AList import AList
from .AMap import AMap
from .BridgeFunction import BridgeFunction
from .BridgeViewer import BridgeViewer
from .Constants import Constants
from .Error import Error
from .Promise import Promise
from .SDKWebSocketServer import SDKWebSocketServer
from .SerializeUtil import SerializeUtil

class SDKViewer3DM(BridgeFunction):

    # the number of function calls made (used to generate call-ids)
    _CALL_COUNT: int = 0
    # the map of function call responses (they arrive from the websocket server) [call-id,serialized-response]
    _RESPONSES: Dict[str,str] = {}

    # Create an instance of the 3DMapping Viewer
    #
    def __init__(self,webSocketServer: SDKWebSocketServer,applicationName: str,options: AMap[object]):
        self.webSocketServer = webSocketServer
        self.applicationName = applicationName
        self.options = options
        self.bridgeViewer = BridgeViewer(applicationName,options,self)
        self.ready = False

    # Check if the javascript viewer if ready
    #
    def isReady(self) -> bool:
        return self.ready

    # Clear the javascript viewer is-ready flag
    #
    def clearReady(self) -> None:
        self.ready = False

    # Get the BridgeViewer
    #
    def getBridgeViewer(self) -> BridgeViewer:
        return self.bridgeViewer

    # Called when the result of a function call arrived
    #
    def setFunctionCallResult(self,callId: str,serializedValue: str) -> None:
        SDKViewer3DM._RESPONSES[callId] = serializedValue

    # Called when a promise result needs to be set
    #
    def setPromiseResult(self,promiseId: int,serializedValue: str) -> None:
        Message.print(__name__,"Got promise "+str(promiseId)+" result")
        value = None
        if (serializedValue!=None) and (len(serializedValue)>0): value = SerializeUtil.deserialize(serializedValue)
        promise = Promise.popInstance(promiseId)
        if (promise!=None):
            Message.print(__name__,"Setting promise "+str(promiseId)+" result")
            promise.setResolved(value)
        if (promise==None):
            Promise.storeResult(promiseId,value)
            Message.print(__name__,"Stored promise "+str(promiseId)+" result")

    # Called when a promise fail result needs to be set
    #
    def setPromiseFailResult(self,promiseId: int,serializedValue: str) -> None:
        Message.print(__name__,"Got promise "+str(promiseId)+" fail result")
        value = None
        if (serializedValue!=None) and (len(serializedValue)>0): value = SerializeUtil.deserialize(serializedValue)
        promise = Promise.popInstance(promiseId)
        if (promise!=None):
            Message.print(__name__,"Setting promise "+str(promiseId)+" fail result")
            promise.setFailed(value)
        if (promise==None):
            Promise.storeFailResult(promiseId,value)
            Message.print(__name__,"Stored promise "+str(promiseId)+" fail result")
    
    # Called when a signal needs dispatching
    #
    def dispatchSignal(self,signalName: str, serializedValue: str) -> None:
        # deserialize the value
        value = None
        if (serializedValue!=None) and (len(serializedValue)>0): value = SerializeUtil.deserialize(serializedValue)
        # dispatch
        if (signalName=="onAppStateChanged"): self.bridgeViewer.onAppStateChanged.dispatchValue(value)
        if (signalName=="onFootprintsChanged"): self.bridgeViewer.onFootprintsChanged.dispatchValue(value)
        if (signalName=="onFieldOfViewsChanged"): self.bridgeViewer.onFieldOfViewsChanged.dispatchValue(value)
        if (signalName=="onGlobalsChanged"): self.bridgeViewer.onGlobalsChanged.dispatchValue(value)
        if (signalName=="onMeasureStateChanged"): self.bridgeViewer.onMeasureStateChanged.dispatchValue(value)
        if (signalName=="onZoomToExtent"): self.bridgeViewer.onZoomToExtent.dispatchValue(value)
        if (signalName=="onFocusPositionChanged"): self.bridgeViewer.onFocusPositionChanged.dispatchValue(value)
        if (signalName=="onPickPositionRequested"): self.bridgeViewer.onPickPositionRequested.dispatchValue(value)
        if (signalName=="onPickPositionCancelled"): self.bridgeViewer.onPickPositionCancelled.dispatchValue(value)
    
    # Process an incoming websocket command (this method gets called by the websocket server)
    #
    def processWebSocketCommand(self,command: str) -> str:
        #
        if (command==None):
            Message.print(__name__,"Got empty command")
            return None
        #
        if (command.startswith("init/")):
            version = command[5:]
            Message.print(__name__,"Got init '"+version+"'")
            self.ready = True
            return "init/"+Constants.SDK_VERSION
        #
        if (command=="getApplicationName"):
            Message.print(__name__,"Got 'getApplicationName' command")
            return "setApplicationName/"+self.applicationName
        #
        if (command=="getStartupOptions"):
            Message.print(__name__,"Got 'getStartupOptions' command")
            serialOptions = SerializeUtil.serialize(self.options)
            return "setStartupOptions/"+serialOptions
        #
        if (command.startswith("dispatchSignal/")):
            try:
                splitIndex = command.find('/',15)
                signalName = command[15:splitIndex]
                serializedValue = command[splitIndex+1:]
                Message.print(__name__,"Got dispatchSignal '"+signalName+"'")
                self.dispatchSignal(signalName,serializedValue)
            except Exception as e:
                Message.printError(__name__,e,"Error while dispatching signal '"+signalName+"'")
            return None
        #
        if (command.startswith("setFunctionCallResult/")):
            splitIndex = command.find('/',22)
            callId = command[22:splitIndex]
            serializedValue = command[splitIndex+1:]
            Message.print(__name__,"Got setFunctionCallResult ["+callId+"]")
            self.setFunctionCallResult(callId,serializedValue)
            return None
        #
        if (command.startswith("setPromiseResult/")):
            try:
                splitIndex = command.find('/',17)
                promiseId = command[17:splitIndex]
                serializedValue = command[splitIndex+1:]
                Message.print(__name__,"Got setPromiseResult ["+promiseId+"]")
                self.setPromiseResult(int(promiseId),serializedValue)
            except Exception as e:
                Message.printError(__name__,e,"Error while setting promise result")
            return None
        #
        if (command.startswith("setPromiseFailResult/")):
            try:
                splitIndex = command.find('/',21)
                promiseId = command[21:splitIndex]
                serializedValue = command[splitIndex+1:]
                Message.print(__name__,"Got setPromiseFailResult ["+promiseId+"] ("+serializedValue+")")
                self.setPromiseFailResult(int(promiseId),serializedValue)
            except Exception as e:
                Message.printError(__name__,e,"Error while setting promise fail result")
            return None
        #
        Message.print(__name__,"Got unknown command '"+command+"'")
        return None

    # Wait for a function call response (this blocks the caller thread for up to 30 seconds)
    #
    def waitForResponse(self,functionName: str,callId: str) -> object:
        iterations = 0
        while True:
            if callId in SDKViewer3DM._RESPONSES:
                return SDKViewer3DM._RESPONSES.pop(callId)
            time.sleep(0.1)
            iterations += 1
            if (iterations==10):
                Message.print(__name__,"waiting for function call "+callId+" response...")
            if (iterations==150):
                Message.print(__name__,"exceeded maximum response wait time for function call id "+callId+", aborting wait")
                raise Exception(__name__+" : Exceeded maximum response wait time for function call id "+callId+", function '"+functionName+"'")
    
    # Call a viewer function and returns the result (blocking call)
    #
    def callViewerFunction(self,functionName: str,arguments: str,waitForResponse: bool) -> str:
        SDKViewer3DM._CALL_COUNT += 1
        callId = str(SDKViewer3DM._CALL_COUNT)
        Message.print(__name__,"calling viewer function '"+functionName+"' with call id "+callId+", waiting? "+str(waitForResponse))
        command = "callFunction/"+callId+"/"+functionName+"/"+arguments
        self.webSocketServer.sendCommandToClient(command)
        response = None
        if (waitForResponse): response = self.waitForResponse(functionName,callId)
        return response

    # BridgeFunction interface method
    #
    def callFunction(self,functionName: str,arguments: AList[object],waitForResponse: bool) -> object:
        serialArguments = SerializeUtil.serialize(arguments)
        serialResult = self.callViewerFunction(functionName,serialArguments,waitForResponse)
        if (serialResult==None): return None
        resultObject = SerializeUtil.deserialize(serialResult)
        if (isinstance(resultObject,Error)): raise Exception("Got javascript error: "+str(resultObject))
        return resultObject
