import asyncio
import threading
import websockets
from typing import Callable
from .....com.orbitgis.tool.message.Message import Message

class SDKWebSocketServer:
    # the counter of client connections
    _CONNECTION_COUNT: int = 0
    
    # Create a new websocket server
    #
    def __init__(self,lostConnectionHandler: Callable[[],None]):
        self.lostConnectionHandler = lostConnectionHandler
        self.commandHandler = None # Callable[[str],str]
        self.webSocket = None
        self.webSocketId = 0
    
    # Check if the websocket client is connected
    #
    def isConnected(self) -> bool:
        return (self.webSocket!=None)
    
    # Set the command handler
    #
    def setCommandHandler(self,handler: Callable[[str],str]) -> None:
        self.commandHandler = handler

    # Send a command from the server (python) to client (javascript)
    #
    def sendCommandToClient(self,command: str) -> None:
        if (self.webSocket==None):
            Message.print(__name__,"No websocket to send command '"+command+"'")
        if (self.webSocket!=None):
            Message.print(__name__,"Sending websocket command '"+command+"' ("+str(self.webSocketId)+")")
            asyncio.run_coroutine_threadsafe(self.webSocket.send(command),self.eventLoop)
    
    # Received a command from the client (javascript) to the server (python)
    #
    def receivedCommandFromClient(self,command: str) -> str:
        if (self.commandHandler==None): return None
        try:
            return self.commandHandler(command)
        except Exception as e:
            Message.printError(__name__,e,"Error while processing websocket command '"+command+"'")
            return None

    # Handle a websocket conversation with a client
    #
    async def connectionHandler(self,awebSocket,path):
        SDKWebSocketServer._CONNECTION_COUNT += 1
        connectionId = SDKWebSocketServer._CONNECTION_COUNT
        self.webSocket = awebSocket
        self.webSocketId = connectionId
        Message.print(__name__,"Received websocket connection ("+str(connectionId)+")")
        while True:
            try:
                #Message.print(__name__,"Awaiting websocket command...")
                command = await self.webSocket.recv()
                commandDisplay = command
                if (len(command)>100): commandDisplay = command[0:100]+"..."
                Message.print(__name__,"Got websocket command: '"+commandDisplay+"' size "+str(len(command))+" ("+str(connectionId)+")")
                if (command=="close"):
                    Message.print(__name__,"Closed client connection ("+str(connectionId)+")")
                    self.webSocket = None
                    return
                response = self.receivedCommandFromClient(command)
                if (response!=None):
                    Message.print(__name__,"Sending websocket response '"+response+"' ("+str(connectionId)+")")
                    await self.webSocket.send(response)
            except Exception as e:
                Message.printError(__name__,e,"Error in websocket server loop, aborting connection") # happens when browser window is closed by the user
                self.webSocket = None
                if (self.lostConnectionHandler!=None):
                    Message.print(__name__,"Calling the lost-connection handler")
                    self.lostConnectionHandler()
                return

    # Start the websocket server in the current thread (blocking)
    #
    def startServerInline(self,hostName: str,port: int) -> None:
        Message.print(__name__,"Starting websocket server on host '"+hostName+"' port "+str(port))
        self.eventLoop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.eventLoop)
        startServerTask = websockets.serve(self.connectionHandler,host=hostName,port=port,max_size=32*1024*1024) # max 32MB messages to receive from client
        self.eventLoop.run_until_complete(startServerTask)
        self.eventLoop.run_forever()
        Message.print(__name__,"Stopped running websocket server")

    # Start the websocket server in a background thread (non-blocking)
    #
    def startServerThread(self,hostName: str,port: int) -> None:
        Message.print(__name__,"Starting websocket server in background thread")
        self.serverThread = threading.Thread(target=self.startServerInline,args=(hostName,port))
        self.serverThread.start()
