import time
from PyQt5.QtCore import QVariant
from qgis.core import QgsApplication,QgsCoordinateReferenceSystem,QgsCoordinateTransform,QgsFeature,QgsField,QgsGeometry,QgsMapLayer,QgsPointXY,QgsProject,QgsTask,QgsVectorLayer
from qgis.gui import QgisInterface,QgsMapCanvas
from ..com.orbitgis.tool.message.Message import Message
from ..orbitgt.mapping3d.sdk.viewer.AList import AList
from ..orbitgt.mapping3d.sdk.viewer.Bounds import Bounds
from ..orbitgt.mapping3d.sdk.viewer.MapObject import MapObject
from ..orbitgt.mapping3d.sdk.viewer.Viewer3DM import Viewer3DM
from .MapObjectLayer import MapObjectLayer
from .MapObjectRenderer import MapObjectRenderer
from .MapObjects import MapObjects

# Class UpdateFootprint manages the updates to the footprint layer
#
class UpdateFootprint:

    # Create a new updater
    #
    def __init__(self,viewer: Viewer3DM,layer: MapObjectLayer,iface: QgisInterface):
        self.viewer = viewer
        self.layer = layer
        self.iface = iface
        self.mapCanvasUpdateTime = 0.0
        self.lastRequestTime = 0.0
        self.isUpdating = False
        self.mapObjects = None

    # Called when the mapcanvas has changed and the footprints need to be reloaded
    #
    def updatedMapCanvas(self):
        self.mapCanvasUpdateTime = time.time()

    # Check for updates
    #
    def checkForUpdates(self) -> None:
        # already updating?
        if (self.isUpdating):
            return
        # already up-to-date?
        if (self.lastRequestTime>self.mapCanvasUpdateTime):
            return
        currentTime = time.time()
        # throttle the rate of update requests
        if (self.lastRequestTime>0.0) and (currentTime<self.lastRequestTime+1.0):
            return
        # throttle the rate of mapcanvas changes (fast zoomin operation)
        if (self.mapCanvasUpdateTime>0.0) and (currentTime<self.mapCanvasUpdateTime+1.0):
            return
        # start the update task (to free the calling timer thread)
        self.isUpdating = True
        self.lastRequestTime = currentTime
        self.updateTask = QgsTask.fromFunction("Update footprints",self.update1,on_finished=self.update2)
        QgsApplication.taskManager().addTask(self.updateTask)

    # Update (in background thread)
    #
    def update1(self,task: QgsTask):
        try:
            Message.print(__name__,"Getting map extent")
            mapExtent = self.iface.mapCanvas().extent() # QgsRectangle
            mapToWorld = QgsCoordinateTransform(self.iface.mapCanvas().mapSettings().destinationCrs(),QgsCoordinateReferenceSystem("EPSG:"+MapObjects.CRS),QgsProject.instance())
            mapScale = self.iface.mapCanvas().scale() # denominator
            p0 = mapToWorld.transform(QgsPointXY(mapExtent.xMinimum(),mapExtent.yMinimum()))
            p1 = mapToWorld.transform(QgsPointXY(mapExtent.xMaximum(),mapExtent.yMinimum()))
            p2 = mapToWorld.transform(QgsPointXY(mapExtent.xMinimum(),mapExtent.yMaximum()))
            p3 = mapToWorld.transform(QgsPointXY(mapExtent.xMaximum(),mapExtent.yMaximum()))
            worldExtent = Bounds(MapObjects.CRS)
            worldExtent.addXY(p0.x(),p0.y())
            worldExtent.addXY(p1.x(),p1.y())
            worldExtent.addXY(p2.x(),p2.y())
            worldExtent.addXY(p3.x(),p3.y())
            Message.print(__name__,"Requesting streaming footprints: bbox "+worldExtent.toString()+" scale "+str(mapScale))
            mapObjectsPromise = self.viewer.getFootprints2D(worldExtent,mapScale)
            Message.print(__name__,"Awaiting footprints with promise id "+str(mapObjectsPromise.getId()))
            self.mapObjects = mapObjectsPromise.waitFor()
            Message.print(__name__,"Got "+str(self.mapObjects.size())+" mapobjects")
            self.isUpdating = False
            return
        except Exception as e:
            Message.printError(__name__,e,"Error while updating footprints")
            self.isUpdating = False
            return
    
    # Finish the update (called by the main UI thread)
    #
    def update2(self,taskException):
        self.layer.update(self.mapObjects,True)
